"Stand-alone program to take PC/Beagle PLUM o/p and produce Python2 routines."

##  PLUM2PY :
##
##  Post-Beagle program, turns PLUM rules into Python2.
##  Copyright 2010, R.S. Forsyth.
##  first version : 20/01/2010
##  last revision : 22/01/2010

import math, os, random, re
import sys, time

NL = '\n'


def main():
    "Main line."

    ##  First some odds & ends concerned with orientation :
    
    args = sys.argv[:]          ##  gets command-line arguments
    na = len(args)
    progname = args[0]
    dateline = time.asctime(time.localtime(time.time()))
    print progname,"[version 1.1]",dateline
    print "command-line args. =",na
    prepath = os.path.dirname(progname)  ##  directory from which program runs
    prelist = ['', prepath]
    print "prepath :",prepath
    if len(sys.path) > 1:
        prelist.append(sys.path[1])
    whereat = os.getcwd()
    print "working folder: ",whereat
    prelist.extend(["a:","C:",sys.path[-1]])
    prelist.append(whereat)
    print "prefix list :",prelist
    
    if na < 2:
        print "script usage:  python " + progname + " <parafile>"
        args.append(raw_input("please give PLUM (Pascal) file name : "))
    plumname = args[1]
    if not plumname.endswith(".pas"):
        print "Program expects .pas file extension."
        plumname = plumname + ".pas"
    me = sys.argv[0]
    print me, plumname
    prepaths = ["C:\\pf"]
    prepaths.append(os.path.dirname(me))

    pasfile = open(plumname,'r')
    print os.path.getsize(plumname)
    ##  also need to get tagname :
    tagname = plumname.replace(".pas",".tag")
    tagname = whereat + '\\' + tagname
    print tagname
    pyprompt = "Please give tag-file name [" + tagname + "] : "
    fromuser = raw_input(pyprompt)
    print len(fromuser)
    if len(fromuser) > 1:  ##  assume it is valid
        tagname = fromuser
    tagfile = open(tagname,'r')
    
    py2name = plumname.replace(".pas",".py")
    py2name = whereat + '\\' + py2name
    print py2name
    pyprompt = "Please give file name for Python2 output [" + py2name + "] : "
    fromuser = raw_input(pyprompt)
    print len(fromuser)
    if len(fromuser) > 1:  ##  assume it is valid
        py2name = fromuser    
    py2file = open(py2name,'w')
    
    ##  heading :
    py2file.write("##  " + dateline + NL)
    py2file.write("##  py2plum output from PC/Beagle rules:")
    py2file.write(NL)
    
    ##  first copy boilerplate stuff from self :
    myself = open(me,'r')
    mylines = myself.readlines()
    copying = 0
    for progline in mylines:
        if progline.startswith("##  START OF BOILER"):
            copying = 1
        if copying:
            py2file.write(progline)
        if progline.startswith("##  END OF BOILER"):
            copying = 0
    myself.close()
    
    ##  next read tagfile :
    taglines = tagfile.readlines()
    tagdict = gettags(taglines)
    ##  print tagdict
    tagfile.close()
    
    ##  next read plum file :
    paslines = pasfile.readlines()
    rulelist = getrules(paslines)
    for r in rulelist:
        print r
    print
    
    ##  check which vars are actually used in rules :
    varsused = tagcheck(paslines,tagdict)
    print varsused
    nv = len(varsused)

    ##  dump rules onto file in python format :
    nr = ruledump(py2file,rulelist,varsused,tagdict)
    print nr, "rules in python output on ",py2name
    print nv, "variables used in rules."
    ##  myself.close()

    ##  preserve some extra info as comments :
    putrems(py2file,paslines)
    print
    endtime = time.asctime(time.localtime(time.time()))
    print "Finish time :",endtime
    py2file.close()
    print "Phoenix is the sweetest lovezums pie! ;-)"
    return


##  RULEDUMP :
def ruledump (fo,rulelist,tagsused,tagdict):
    "dumps rules onto output file as python function."

    fo.write(NL)
    fo.write("##  BEAGVAL :\n")
    fo.write("def beagval (varvals):\n")
    fo.write('    "evaluates Beagle rules on list of values (first at 0)."\n')
    fo.write("    ##  take care to supply values in same positions as tags in tagfile.\n")
    fo.write(NL)
    tags = tagdict.keys()
    nt = len(tags)
    for t in tags:
        lout = "    " + t + " = varvals[" + str(tagdict[t]) + "]"
        if t in tagsused:
            fo.write(lout + NL)
    fo.write("    Z = 0.0  ##  zero variable\n")
    fo.write(NL)
    
    nr = len(rulelist)
    fo.write("    rtot = 0\n")
    for r in rulelist:
        lout = "    rv = " + r
        fo.write(lout + NL)  ##  should be fully bracketted
        lout = "    if rv:"
        fo.write(lout + NL)  ##  PLUM Pascal rules always boolean at top level
        fo.write("        rtot += 1\n")

    fo.write(NL)
    fo.write("    if rtot >= " + str(nr) + ":\n")
    fo.write("        return  'all'\n")
    fo.write("    if rtot <= 0:\n")
    fo.write("        return  'none'\n")
    fo.write("    return  'some'\n" + NL)
    
    return  nr


##  PUTREMS :
def putrems (fo,paslines):
    "puts various comment lines onto python o/p file for reference."

    lc = 0
    sincetv = 0
    for pasline in paslines:
        lc += 1
        line = pasline.strip()
        if line.startswith("{ Coded from "):
            fo.write("##  " + line + NL)
        if line.startswith("{ Cannot be reached"):
            tp = lc

    targline = paslines[tp+1].strip()
    if targline.endswith(';'):
        targline = targline[:-1]
    fo.write("##  Beagle target expression : " + targline + NL)

    fo.write(NL)
    fo.write("##  function takes input data case (file row) in varvals[].\n")
    fo.write("##  returns 'none', 'some' or 'all' based on # of rules firing.\n")

    return


##  TAGCHECK :
def tagcheck (paslines,tagdict):
    "finds which tags are actually used in rules."

    looking = 0
    usedlist = []
    for pasline in paslines:
        line = pasline.strip()
        if line.startswith("Procedure Beagle ("):
            looking = 1
        if line.startswith("{ N.B. tag"):
            looking = 0
            break
        if looking and not line.startswith("Procedure Beagle"):
            bits = line.split(',')
            for b in bits:
                bx = b.strip()
                if bx and bx.find(':') < 0:  ##  mustn't contain colon
                    usedlist.append(bx)

    return  usedlist


##  GETRULES :
def getrules (paslines):
    "extracts rules from Pascal lines produced by PC/Beagle PLUM."

    lc = 0
    rulelist = []
    thisrule = ''
    for pasline in paslines:
        line = pasline.strip()
        if line.startswith("rule["):  ##  start of new rule
            r = line.split(":=")
            rule = r[1]
            if thisrule:
                rulelist.append(thisrule)
            thisrule = rule
        else:
            if thisrule and not line.startswith("ipos :="):
                thisrule = thisrule + ' ' + line
        if line.startswith("ipos :="):  ##  enough read
            if thisrule:
                rulelist.append(thisrule)
            break

    ##  tidy up rule list :
    pyrules = []
    for r in rulelist:
        rx = r.replace("bool(","beagbool(")
        rx = rx.replace("math(","beagmath(")
        rx = rx.replace(" AND "," and ")
        rx = rx.replace(" OR "," or ")
        rx = rx.replace("NOT ","not ")
        pyrules.append(rx[:-1])  ##  last char always semi-colon
        
    return  pyrules

        
##  GETTAGS :
def gettags (taglines,sepchar=','):
    "makes variable-name table from tagfile lines."

    tagdict = {}
    for tp in range(len(taglines)):
        bits = taglines[tp].split(sepchar)  ##  always comma-delimited
        vartext = bits[0].strip().upper()  ##  Beagle folds names to upper case
        varname = ''
        ##  ignore non-alphanumerics that might be included, as does Beagle :
        for ch in vartext:
            if ch.isalnum():
                varname = varname + ch
        if len(varname) > 0:
            tagdict[varname] = tp  ##  record position

    return  tagdict


##  START OF BOILERPLATE.

##  BEAGBOOL :
def beagbool (v):
    "Python function to convert numeric to Boolean, as in Beagle."

    if v > 0:
        return  True
    else:
        return  False


##  BEAGMATH :
def beagmath (v):
    "Python function to convert Boolean to numeric, as in Beagle."

    if v:
        return  1.0
    else:
        return  0.0
    

##  END OF BOILERPLATE.


##  MAKEROWS :
def makerows (linelist,rexlist,minsize=2):
    "makes Beagle-style data rows from lines read from CL08-style file."

    rowlist = []
    lc = 0 ; rc = 0
    linelab = "Zz"
    lastline = '' ; lout = ''
    for datline in linelist:
        line = datline.strip()
        linesize = len(line)
        if linesize < minsize:
            continue  ##  skip if too short
        if line.startswith(Premark):  ##  coding data
            bits = line[2:].split('=')
            for j in range(len(bits)):
                bits[j] = bits[j].strip()
                attname = bits[0] ; attvalue = bits[1]
            if attname == "varvals":
                ##  recompute values from text line :
                varlist = par2vars(lastline,rexlist)
                attvalue = str(varlist)
                lout = lout + ', ' + attvalue[1:-1]
            if attname == "talkcat":
                lout = lout + ', ' + attvalue + ', ' + Codenums[attvalue]
                ##  no room for errors!
                rowlist.append(lout)
                lout = ''
                ##  talkcat must come last.
            if rc < 8:
                print linelab,attname,attvalue
                print lout
        else:  ##  compose label from text line
            lc += 1
            linelab = makelab(line,lc)
            lout = linelab
            lastline = line
            ##  assumes Premark lines preceded by paraline.
        rc += 1

    return  rowlist


##  MAKELAB :
def makelab (textline,rc):
    "makes a line label acceptable as Beagle input."

    line = textline[:16]
    lout = "para"+str(rc)
    for ch in line:
        if ch.isalnum():
            lout = lout + ch

    return  lout


##  REXFIND :
def rexfind (rexlist,textline):
    "finds number of occurrences of rexlist's regexes in textline."

    rexfrex = []
    for rx in rexlist:
        hitlist = rx.findall(textline)
        hits = len(hitlist)
        rexfrex.append(hits)

    return  rexfrex


##  PATCOMP :
def patcomp (patlist):
    "compiles list of pattern strings into a list of regex objects."

    rexlist = []
    for p in patlist:
        rx = re.compile(p)
        rexlist.append(rx)

    return  rexlist


##  GETPATS :
def getpats (patfile):
    "reads (re-style) pattern strings from file."

    patlist = []
    linelist = patfile.readlines()
    for patt in linelist:
        p = patt.strip()
        if len(p) > 0:
            patlist.append(p)  ##  no checking

    return  patlist


##  QUOTEPOS :
def quotepos (paraline):
    "returns first & last position of quotes if found in paraline."

    pl = paraline.strip()
    nc = len(pl)
    q1 = -1
    for p in range(nc):
        if pl[p] in Anyquote:
            q1 = p
            break
    qx = -1
    for p in range(nc-1,0,-1):
        if pl[p] in Anyquote:
            qx = p
            break

    return (q1,qx)


##  PARAGRAB :
def paragrab (linelist):
    "takes a list of text lines and returns a list of paragraphs."

    paralist = []
    thispar = ''
    for datline in linelist:
        line = datline.strip()
        nc = len(line)
        if nc < 1:  ##  blank taken as paragraph ending
            if len(thispar) > 0:
                paralist.append(thispar.strip())
            thispar = ''
        else:  ##  within para (single cr/lf)
            thispar = thispar + datline

    return  paralist




main()
